@@ -1,85 +1,59 @@
 #! /usr/bin/env python
 
+__copyright__ = "Copyright (C) 2009 Andreas Kloeckner"
+
+__license__ = """
+Permission is hereby granted, free of charge, to any person obtaining a copy
+of this software and associated documentation files (the "Software"), to deal
+in the Software without restriction, including without limitation the rights
+to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
+copies of the Software, and to permit persons to whom the Software is
+furnished to do so, subject to the following conditions:
+
+The above copyright notice and this permission notice shall be included in
+all copies or substantial portions of the Software.
+
+THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
+THE SOFTWARE.
+"""
+
 import numpy as np
 import numpy.linalg as la
 import sys
 import pytools.test
 
-
-def have_cl():
-    try:
-        import pyopencl
-        return True
-    except:
-        return False
-
-if have_cl():
-    import pyopencl as cl
-    import pyopencl.array as cl_array
-    import pyopencl.tools as cl_tools
-    from pyopencl.tools import pytest_generate_tests_for_pyopencl \
-            as pytest_generate_tests
-    from pyopencl.characterize import has_double_support
+import pyopencl as cl
+import pyopencl.array as cl_array
+import pyopencl.tools as cl_tools
+from pyopencl.tools import (  # noqa
+        pytest_generate_tests_for_pyopencl as pytest_generate_tests)
+from pyopencl.characterize import has_double_support
 
 
-@pytools.test.mark_test.opencl
-def test_pow_array(ctx_factory):
-    context = ctx_factory()
-    queue = cl.CommandQueue(context)
-
-    a = np.array([1, 2, 3, 4, 5]).astype(np.float32)
-    a_gpu = cl_array.to_device(queue, a)
-
-    result = pow(a_gpu, a_gpu).get()
-    assert (np.abs(a ** a - result) < 1e-3).all()
-
-    result = (a_gpu ** a_gpu).get()
-    assert (np.abs(pow(a, a) - result) < 1e-3).all()
-
-
-@pytools.test.mark_test.opencl
-def test_pow_number(ctx_factory):
-    context = ctx_factory()
-    queue = cl.CommandQueue(context)
-
-    a = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]).astype(np.float32)
-    a_gpu = cl_array.to_device(queue, a)
-
-    result = pow(a_gpu, 2).get()
-    assert (np.abs(a ** 2 - result) < 1e-3).all()
-
-
-@pytools.test.mark_test.opencl
-def test_absrealimag(ctx_factory):
-    context = ctx_factory()
-    queue = cl.CommandQueue(context)
-
-    def real(x): return x.real
-    def imag(x): return x.imag
-    def conj(x): return x.conj()
-
-    n = 111
-    for func in [abs, real, imag, conj]:
-        for dtype in [np.int32, np.float32, np.complex64]:
-            print(func, dtype)
-            a = -make_random_array(queue, dtype, n)
-
-            host_res = func(a.get())
-            dev_res = func(a).get()
-
-            correct = np.allclose(dev_res, host_res)
-            if not correct:
-                print(dev_res)
-                print(host_res)
-                print(dev_res-host_res)
-            assert correct
-
+# {{{ helpers
 
 TO_REAL = {
         np.dtype(np.complex64): np.float32,
         np.dtype(np.complex128): np.float64
         }
 
+
+def general_clrand(queue, shape, dtype):
+    from pyopencl.clrandom import rand as clrand
+
+    dtype = np.dtype(dtype)
+    if dtype.kind == "c":
+        real_dtype = dtype.type(0).real.dtype
+        return clrand(queue, shape, real_dtype) + 1j*clrand(queue, shape, real_dtype)
+    else:
+        return clrand(queue, shape, dtype)
+
+
 def make_random_array(queue, dtype, size):
     from pyopencl.clrandom import rand
 
@@ -87,11 +61,16 @@ def make_random_array(queue, dtype, size):
     if dtype.kind == "c":
         real_dtype = TO_REAL[dtype]
         return (rand(queue, shape=(size,), dtype=real_dtype).astype(dtype)
-                + dtype.type(1j)
-                * rand(queue, shape=(size,), dtype=real_dtype).astype(dtype))
+                + rand(queue, shape=(size,), dtype=real_dtype).astype(dtype)
+                * dtype.type(1j))
     else:
         return rand(queue, shape=(size,), dtype=dtype)
 
+# }}}
+
+
+# {{{ dtype-related
+
 @pytools.test.mark_test.opencl
 def test_basic_complex(ctx_factory):
     context = ctx_factory()
@@ -101,12 +80,13 @@ def test_basic_complex(ctx_factory):
 
     size = 500
 
-    ary =  (rand(queue, shape=(size,), dtype=np.float32).astype(np.complex64)
-            + 1j* rand(queue, shape=(size,), dtype=np.float32).astype(np.complex64))
+    ary = (rand(queue, shape=(size,), dtype=np.float32).astype(np.complex64)
+            + rand(queue, shape=(size,), dtype=np.float32).astype(np.complex64) * 1j)
     c = np.complex64(5+7j)
 
     host_ary = ary.get()
-    assert la.norm((c*ary).get() - c*host_ary) < 1e-5 * la.norm(host_ary)
+    assert la.norm((ary*c).get() - c*host_ary) < 1e-5 * la.norm(host_ary)
+
 
 @pytools.test.mark_test.opencl
 def test_mix_complex(ctx_factory):
@@ -178,14 +158,115 @@ def test_mix_complex(ctx_factory):
 
                     assert correct
 
+
 @pytools.test.mark_test.opencl
-def test_len(ctx_factory):
+def test_pow_neg1_vs_inv(ctx_factory):
+    ctx = ctx_factory()
+    queue = cl.CommandQueue(ctx)
+
+    device = ctx.devices[0]
+    if not has_double_support(device):
+        from pytest import skip
+        skip("double precision not supported on %s" % device)
+
+    a_dev = make_random_array(queue, np.complex128, 20000)
+
+    res1 = (a_dev ** (-1)).get()
+    res2 = (1/a_dev).get()
+    ref = 1/a_dev.get()
+
+    assert la.norm(res1-ref, np.inf) / la.norm(ref) < 1e-13
+    assert la.norm(res2-ref, np.inf) / la.norm(ref) < 1e-13
+
+
+@pytools.test.mark_test.opencl
+def test_vector_fill(ctx_factory):
+    context = ctx_factory()
+    queue = cl.CommandQueue(context)
+
+    a_gpu = cl_array.Array(queue, 100, dtype=cl_array.vec.float4)
+    a_gpu.fill(cl_array.vec.make_float4(0.0, 0.0, 1.0, 0.0))
+    a = a_gpu.get()
+    assert a.dtype is cl_array.vec.float4
+
+    a_gpu = cl_array.zeros(queue, 100, dtype=cl_array.vec.float4)
+
+
+@pytools.test.mark_test.opencl
+def test_absrealimag(ctx_factory):
+    context = ctx_factory()
+    queue = cl.CommandQueue(context)
+
+    def real(x):
+        return x.real
+
+    def imag(x):
+        return x.imag
+
+    def conj(x):
+        return x.conj()
+
+    n = 111
+    for func in [abs, real, imag, conj]:
+        for dtype in [np.int32, np.float32, np.complex64]:
+            print(func, dtype)
+            a = -make_random_array(queue, dtype, n)
+
+            host_res = func(a.get())
+            dev_res = func(a).get()
+
+            correct = np.allclose(dev_res, host_res)
+            if not correct:
+                print(dev_res)
+                print(host_res)
+                print(dev_res-host_res)
+            assert correct
+
+# }}}
+
+
+# {{{ operators
+
+@pytools.test.mark_test.opencl
+def test_rmul_yields_right_type(ctx_factory):
+    context = ctx_factory()
+    queue = cl.CommandQueue(context)
+
+    a = np.array([1, 2, 3, 4, 5]).astype(np.float32)
+    a_gpu = cl_array.to_device(queue, a)
+
+    two_a = 2*a_gpu
+    assert isinstance(two_a, cl_array.Array)
+
+    two_a = np.float32(2)*a_gpu
+    assert isinstance(two_a, cl_array.Array)
+
+
+@pytools.test.mark_test.opencl
+def test_pow_array(ctx_factory):
+    context = ctx_factory()
+    queue = cl.CommandQueue(context)
+
+    a = np.array([1, 2, 3, 4, 5]).astype(np.float32)
+    a_gpu = cl_array.to_device(queue, a)
+
+    result = pow(a_gpu, a_gpu).get()
+    assert (np.abs(a ** a - result) < 1e-3).all()
+
+    result = (a_gpu ** a_gpu).get()
+    assert (np.abs(pow(a, a) - result) < 1e-3).all()
+
+
+@pytools.test.mark_test.opencl
+def test_pow_number(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
 
     a = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]).astype(np.float32)
-    a_cpu = cl_array.to_device(queue, a)
-    assert len(a_cpu) == 10
+    a_gpu = cl_array.to_device(queue, a)
+
+    result = pow(a_gpu, 2).get()
+    assert (np.abs(a ** 2 - result) < 1e-3).all()
 
 
 @pytools.test.mark_test.opencl
@@ -197,9 +278,9 @@ def test_multiply(ctx_factory):
 
     for sz in [10, 50000]:
         for dtype, scalars in [
-            (np.float32, [2]),
-            (np.complex64, [2j]),
-            ]:
+                (np.float32, [2]),
+                (np.complex64, [2j]),
+                ]:
             for scalar in scalars:
                 a_gpu = make_random_array(queue, dtype, sz)
                 a = a_gpu.get()
@@ -331,6 +412,10 @@ def test_divide_array(ctx_factory):
     a_divide = (b_gpu / a_gpu).get()
     assert (np.abs(b / a - a_divide) < 1e-3).all()
 
+# }}}
+
+
+# {{{ RNG
 
 @pytools.test.mark_test.opencl
 def test_random(ctx_factory):
@@ -371,8 +456,41 @@ def test_random(ctx_factory):
         #pt.hist(ran.get())
         #pt.show()
 
+# }}}
 
 
+# {{{ misc
+
+@pytools.test.mark_test.opencl
+def test_numpy_integer_shape(ctx_factory):
+    context = ctx_factory()
+    queue = cl.CommandQueue(context)
+
+    cl_array.empty(queue, np.int32(17), np.float32)
+    cl_array.empty(queue, (np.int32(17), np.int32(17)), np.float32)
+
+
+@pytools.test.mark_test.opencl
+def test_len(ctx_factory):
+    context = ctx_factory()
+    queue = cl.CommandQueue(context)
+
+    a = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]).astype(np.float32)
+    a_cpu = cl_array.to_device(queue, a)
+    assert len(a_cpu) == 10
+
+
+@pytools.test.mark_test.opencl
+def test_stride_preservation(ctx_factory):
+    context = ctx_factory()
+    queue = cl.CommandQueue(context)
+
+    A = np.random.rand(3, 3)
+    AT = A.T
+    print(AT.flags.f_contiguous, AT.flags.c_contiguous)
+    AT_GPU = cl_array.to_device(queue, AT)
+    print(AT_GPU.flags.f_contiguous, AT_GPU.flags.c_contiguous)
+    assert np.allclose(AT_GPU.get(), AT)
 
 
 @pytools.test.mark_test.opencl
@@ -401,386 +519,18 @@ def test_nan_arithmetic(ctx_factory):
     assert (np.isnan(ab) == np.isnan(ab_gpu)).all()
 
 
-@pytools.test.mark_test.opencl
-def test_elwise_kernel(ctx_factory):
-    context = ctx_factory()
-    queue = cl.CommandQueue(context)
-
-    from pyopencl.clrandom import rand as clrand
-
-    a_gpu = clrand(queue, (50,), np.float32)
-    b_gpu = clrand(queue, (50,), np.float32)
-
-    from pyopencl.elementwise import ElementwiseKernel
-    lin_comb = ElementwiseKernel(context,
-            "float a, float *x, float b, float *y, float *z",
-            "z[i] = a*x[i] + b*y[i]",
-            "linear_combination")
-
-    c_gpu = cl_array.empty_like(a_gpu)
-    lin_comb(5, a_gpu, 6, b_gpu, c_gpu)
-
-    assert la.norm((c_gpu - (5 * a_gpu + 6 * b_gpu)).get()) < 1e-5
-
-
-@pytools.test.mark_test.opencl
-def test_elwise_kernel_with_options(ctx_factory):
-    from pyopencl.clrandom import rand as clrand
-    from pyopencl.elementwise import ElementwiseKernel
-
-    context = ctx_factory()
-    queue = cl.CommandQueue(context)
-
-    in_gpu = clrand(queue, (50,), np.float32)
-
-    options = ['-D', 'ADD_ONE']
-    add_one = ElementwiseKernel(
-        context,
-        "float* out, const float *in",
-        """
-        out[i] = in[i]
-        #ifdef ADD_ONE
-            +1
-        #endif
-        ;
-        """,
-        options=options,
-        )
-
-    out_gpu = cl_array.empty_like(in_gpu)
-    add_one(out_gpu, in_gpu)
-
-    gt = in_gpu.get() + 1
-    gv = out_gpu.get()
-    assert la.norm(gv - gt) < 1e-5
-
-
-@pytools.test.mark_test.opencl
-def test_take(ctx_factory):
-    context = ctx_factory()
-    queue = cl.CommandQueue(context)
-
-    idx = cl_array.arange(queue, 0, 200000, 2, dtype=np.uint32)
-    a = cl_array.arange(queue, 0, 600000, 3, dtype=np.float32)
-    result = cl_array.take(a, idx)
-    assert ((3 * idx).get() == result.get()).all()
-
-
-@pytools.test.mark_test.opencl
-def test_arange(ctx_factory):
-    context = ctx_factory()
-    queue = cl.CommandQueue(context)
-
-    n = 5000
-    a = cl_array.arange(queue, n, dtype=np.float32)
-    assert (np.arange(n, dtype=np.float32) == a.get()).all()
-
-
-@pytools.test.mark_test.opencl
-def test_reverse(ctx_factory):
-    context = ctx_factory()
-    queue = cl.CommandQueue(context)
-
-    n = 5000
-    a = np.arange(n).astype(np.float32)
-    a_gpu = cl_array.to_device(queue, a)
-
-    a_gpu = a_gpu.reverse()
-
-    assert (a[::-1] == a_gpu.get()).all()
-
-
-@pytools.test.mark_test.opencl
-def test_sum(ctx_factory):
-    context = ctx_factory()
-    queue = cl.CommandQueue(context)
-
-    from pyopencl.clrandom import rand as clrand
-
-    a_gpu = clrand(queue, (200000,), np.float32)
-    a = a_gpu.get()
-
-    sum_a = np.sum(a)
-    sum_a_gpu = cl_array.sum(a_gpu).get()
-
-    assert abs(sum_a_gpu - sum_a) / abs(sum_a) < 1e-4
-
-
-@pytools.test.mark_test.opencl
-def test_minmax(ctx_factory):
-    context = ctx_factory()
-    queue = cl.CommandQueue(context)
-
-    from pyopencl.clrandom import rand as clrand
-
-    if has_double_support(context.devices[0]):
-        dtypes = [np.float64, np.float32, np.int32]
-    else:
-        dtypes = [np.float32, np.int32]
-
-    for what in ["min", "max"]:
-        for dtype in dtypes:
-            a_gpu = clrand(queue, (200000,), dtype)
-            a = a_gpu.get()
-
-            op_a = getattr(np, what)(a)
-            op_a_gpu = getattr(cl_array, what)(a_gpu).get()
-
-            assert op_a_gpu == op_a, (op_a_gpu, op_a, dtype, what)
-
-
-@pytools.test.mark_test.opencl
-def test_subset_minmax(ctx_factory):
-    context = ctx_factory()
-    queue = cl.CommandQueue(context)
-
-    from pyopencl.clrandom import rand as clrand
-
-    l_a = 200000
-    gran = 5
-    l_m = l_a - l_a // gran + 1
-
-    if has_double_support(context.devices[0]):
-        dtypes = [np.float64, np.float32, np.int32]
-    else:
-        dtypes = [np.float32, np.int32]
-
-    for dtype in dtypes:
-        a_gpu = clrand(queue, (l_a,), dtype)
-        a = a_gpu.get()
-
-        meaningful_indices_gpu = cl_array.zeros(
-                queue, l_m, dtype=np.int32)
-        meaningful_indices = meaningful_indices_gpu.get()
-        j = 0
-        for i in range(len(meaningful_indices)):
-            meaningful_indices[i] = j
-            j = j + 1
-            if j % gran == 0:
-                j = j + 1
-
-        meaningful_indices_gpu = cl_array.to_device(
-                queue, meaningful_indices)
-        b = a[meaningful_indices]
-
-        min_a = np.min(b)
-        min_a_gpu = cl_array.subset_min(meaningful_indices_gpu, a_gpu).get()
-
-        assert min_a_gpu == min_a
-
-
-@pytools.test.mark_test.opencl
-def test_dot(ctx_factory):
-    context = ctx_factory()
-    queue = cl.CommandQueue(context)
-
-    from pyopencl.clrandom import rand as clrand
-    a_gpu = clrand(queue, (200000,), np.float32)
-    a = a_gpu.get()
-    b_gpu = clrand(queue, (200000,), np.float32)
-    b = b_gpu.get()
-
-    dot_ab = np.dot(a, b)
-
-    dot_ab_gpu = cl_array.dot(a_gpu, b_gpu).get()
-
-    assert abs(dot_ab_gpu - dot_ab) / abs(dot_ab) < 1e-4
-
-
-if False:
-    @pytools.test.mark_test.opencl
-    def test_slice(ctx_factory):
-        from pyopencl.clrandom import rand as clrand
-
-        l = 20000
-        a_gpu = clrand(queue, (l,))
-        a = a_gpu.get()
-
-        from random import randrange
-        for i in range(200):
-            start = randrange(l)
-            end = randrange(start, l)
-
-            a_gpu_slice = a_gpu[start:end]
-            a_slice = a[start:end]
-
-            assert la.norm(a_gpu_slice.get() - a_slice) == 0
-
-
-@pytools.test.mark_test.opencl
-def test_if_positive(ctx_factory):
-    context = ctx_factory()
-    queue = cl.CommandQueue(context)
-
-    from pyopencl.clrandom import rand as clrand
-
-    l = 20000
-    a_gpu = clrand(queue, (l,), np.float32)
-    b_gpu = clrand(queue, (l,), np.float32)
-    a = a_gpu.get()
-    b = b_gpu.get()
-
-    max_a_b_gpu = cl_array.maximum(a_gpu, b_gpu)
-    min_a_b_gpu = cl_array.minimum(a_gpu, b_gpu)
-
-    print(max_a_b_gpu)
-    print(np.maximum(a, b))
-
-    assert la.norm(max_a_b_gpu.get() - np.maximum(a, b)) == 0
-    assert la.norm(min_a_b_gpu.get() - np.minimum(a, b)) == 0
-
-
-@pytools.test.mark_test.opencl
-def test_take_put(ctx_factory):
-    context = ctx_factory()
-    queue = cl.CommandQueue(context)
-
-    for n in [5, 17, 333]:
-        one_field_size = 8
-        buf_gpu = cl_array.zeros(queue,
-                n * one_field_size, dtype=np.float32)
-        dest_indices = cl_array.to_device(queue,
-                np.array([0, 1, 2,  3, 32, 33, 34, 35], dtype=np.uint32))
-        read_map = cl_array.to_device(queue,
-                np.array([7, 6, 5, 4, 3, 2, 1, 0], dtype=np.uint32))
-
-        cl_array.multi_take_put(
-                arrays=[buf_gpu for i in range(n)],
-                dest_indices=dest_indices,
-                src_indices=read_map,
-                src_offsets=[i * one_field_size for i in range(n)],
-                dest_shape=(96,))
-
-
-@pytools.test.mark_test.opencl
-def test_astype(ctx_factory):
-    context = ctx_factory()
-    queue = cl.CommandQueue(context)
-
-    from pyopencl.clrandom import rand as clrand
-
-    if not has_double_support(context.devices[0]):
-        return
-
-    a_gpu = clrand(queue, (2000,), dtype=np.float32)
-
-    a = a_gpu.get().astype(np.float64)
-    a2 = a_gpu.astype(np.float64).get()
-
-    assert a2.dtype == np.float64
-    assert la.norm(a - a2) == 0, (a, a2)
-
-    a_gpu = clrand(queue, (2000,), dtype=np.float64)
-
-    a = a_gpu.get().astype(np.float32)
-    a2 = a_gpu.astype(np.float32).get()
-
-    assert a2.dtype == np.float32
-    assert la.norm(a - a2) / la.norm(a) < 1e-7
-
-
-def summarize_error(obtained, desired, orig, thresh=1e-5):
-    err = obtained - desired
-    ok_count = 0
-
-    entries = []
-    for i, val in enumerate(err):
-        if abs(val) > thresh:
-            if ok_count:
-                entries.append("<%d ok>" % ok_count)
-                ok_count = 0
-
-            entries.append("%r (want: %r, diff: %r, orig: %r)" % (obtained[i], desired[i],
-                obtained[i]-desired[i], orig[i]))
-        else:
-            ok_count += 1
-
-    if ok_count:
-        entries.append("<%d ok>" % ok_count)
-
-    return " ".join(entries)
-
-@pytools.test.mark_test.opencl
-def test_scan(ctx_factory):
-    context = ctx_factory()
-    queue = cl.CommandQueue(context)
-
-    from pyopencl.scan import InclusiveScanKernel, ExclusiveScanKernel
-
-    dtype = np.int32
-    for cls in [
-            InclusiveScanKernel,
-            ExclusiveScanKernel
-            ]:
-        knl = cls(context, dtype, "a+b", "0")
-
-        for n in [
-            10, 2 ** 10 - 5, 2 ** 10,
-            2 ** 20 - 2 ** 18,
-            2 ** 20 - 2 ** 18 + 5,
-            2 ** 10 + 5,
-            2 ** 20 + 1,
-            2 ** 20, 2 ** 24
-            ]:
-
-            host_data = np.random.randint(0, 10, n).astype(dtype)
-            dev_data = cl_array.to_device(queue, host_data)
-
-            assert (host_data == dev_data.get()).all() # /!\ fails on Nv GT2?? for some drivers
-
-            knl(dev_data)
-
-            desired_result = np.cumsum(host_data, axis=0)
-            if cls is ExclusiveScanKernel:
-                desired_result -= host_data
-
-            is_ok = (dev_data.get() == desired_result).all()
-            if 0 and not is_ok:
-                print(summarize_error(dev_data.get(), desired_result, host_data))
-
-            assert is_ok
-            from gc import collect
-            collect()
-
-
-@pytools.test.mark_test.opencl
-def test_stride_preservation(ctx_factory):
-    context = ctx_factory()
-    queue = cl.CommandQueue(context)
-
-    A = np.random.rand(3, 3)
-    AT = A.T
-    print(AT.flags.f_contiguous, AT.flags.c_contiguous)
-    AT_GPU = cl_array.to_device(queue, AT)
-    print(AT_GPU.flags.f_contiguous, AT_GPU.flags.c_contiguous)
-    assert np.allclose(AT_GPU.get(), AT)
-
-
-@pytools.test.mark_test.opencl
-def test_vector_fill(ctx_factory):
-    context = ctx_factory()
-    queue = cl.CommandQueue(context)
-
-    a_gpu = cl_array.Array(queue, 100, dtype=cl_array.vec.float4)
-    a_gpu.fill(cl_array.vec.make_float4(0.0, 0.0, 1.0, 0.0))
-    a = a_gpu.get()
-    assert a.dtype is cl_array.vec.float4
-
-    a_gpu = cl_array.zeros(queue, 100, dtype=cl_array.vec.float4)
-
 @pytools.test.mark_test.opencl
 def test_mem_pool_with_arrays(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
-    mem_pool = cl_tools.MemoryPool(cl_tools.CLAllocator(context))
+    mem_pool = cl_tools.MemoryPool(cl_tools.ImmediateAllocator(queue))
 
     a_dev = cl_array.arange(queue, 2000, dtype=np.float32, allocator=mem_pool)
     b_dev = cl_array.to_device(queue, np.arange(2000), allocator=mem_pool) + 4000
 
-    result = cl_array.dot(a_dev, b_dev)
     assert a_dev.allocator is mem_pool
     assert b_dev.allocator is mem_pool
-    assert result.allocator is mem_pool
+
 
 @pytools.test.mark_test.opencl
 def test_view(ctx_factory):
@@ -802,92 +552,78 @@ def test_view(ctx_factory):
     view = a_dev.view(np.int16)
     assert view.shape == (8, 32) and view.dtype == np.int16
 
-mmc_dtype = np.dtype([
-    ("cur_min", np.int32),
-    ("cur_max", np.int32),
-    ("pad", np.int32),
-    ])
+# }}}
 
-from pyopencl.tools import register_dtype
-register_dtype(mmc_dtype, "minmax_collector", alias_ok=True)
-register_dtype(mmc_dtype, "minmax_collector", alias_ok=True)
 
 @pytools.test.mark_test.opencl
-def test_struct_reduce(ctx_factory):
+def test_slice(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
 
-    preamble = r"""//CL//
-    struct minmax_collector
-    {
-        int cur_min;
-        int cur_max;
-        // Workaround for OS X Lion GPU CL. Mystifying.
-        int pad;
-    };
+    from pyopencl.clrandom import rand as clrand
 
-    typedef struct minmax_collector minmax_collector;
+    l = 20000
+    a_gpu = clrand(queue, (l,), dtype=np.float32)
+    b_gpu = clrand(queue, (l,), dtype=np.float32)
+    a = a_gpu.get()
+    b = b_gpu.get()
 
-    minmax_collector mmc_neutral()
-    {
-        // FIXME: needs infinity literal in real use, ok here
-        minmax_collector result;
-        result.cur_min = 1<<30;
-        result.cur_max = -(1<<30);
-        return result;
-    }
+    from random import randrange
+    for i in range(20):
+        start = randrange(l)
+        end = randrange(start, l)
 
-    minmax_collector mmc_from_scalar(float x)
-    {
-        minmax_collector result;
-        result.cur_min = x;
-        result.cur_max = x;
-        return result;
-    }
+        a_gpu_slice = 2*a_gpu[start:end]
+        a_slice = 2*a[start:end]
 
-    minmax_collector agg_mmc(minmax_collector a, minmax_collector b)
-    {
-        minmax_collector result = a;
-        if (b.cur_min < result.cur_min)
-            result.cur_min = b.cur_min;
-        if (b.cur_max > result.cur_max)
-            result.cur_max = b.cur_max;
-        return result;
-    }
+        assert la.norm(a_gpu_slice.get() - a_slice) == 0
 
-    """
+    for i in range(20):
+        start = randrange(l)
+        end = randrange(start, l)
 
+        a_gpu[start:end] = 2*b[start:end]
+        a[start:end] = 2*b[start:end]
+
+        assert la.norm(a_gpu.get() - a) == 0
+
+    for i in range(20):
+        start = randrange(l)
+        end = randrange(start, l)
+
+        a_gpu[start:end] = 2*b_gpu[start:end]
+        a[start:end] = 2*b[start:end]
+
+        assert la.norm(a_gpu.get() - a) == 0
+
+
+@pytools.test.mark_test.opencl
+def test_concatenate(ctx_factory):
+    context = ctx_factory()
+    queue = cl.CommandQueue(context)
 
     from pyopencl.clrandom import rand as clrand
-    a_gpu = clrand(queue, (20000,), dtype=np.int32, a=0, b=10**6)
-    a = a_gpu.get()
 
-    from pyopencl.reduction import ReductionKernel
-    red = ReductionKernel(context, mmc_dtype,
-            neutral="mmc_neutral()",
-            reduce_expr="agg_mmc(a, b)", map_expr="mmc_from_scalar(x[i])",
-            arguments="__global int *x", preamble=preamble)
-
-    minmax = red(a_gpu).get()
-    #print minmax["cur_min"], minmax["cur_max"]
-    #print np.min(a), np.max(a)
-
-    assert abs(minmax["cur_min"] - np.min(a)) < 1e-5
-    assert abs(minmax["cur_max"] - np.max(a)) < 1e-5
+    a_dev = clrand(queue, (5, 15, 20), dtype=np.float32)
+    b_dev = clrand(queue, (4, 15, 20), dtype=np.float32)
+    c_dev = clrand(queue, (3, 15, 20), dtype=np.float32)
+    a = a_dev.get()
+    b = b_dev.get()
+    c = c_dev.get()
 
+    cat_dev = cl.array.concatenate((a_dev, b_dev, c_dev))
+    cat = np.concatenate((a, b, c))
 
+    assert la.norm(cat - cat_dev.get()) == 0
 
 
 if __name__ == "__main__":
     # make sure that import failures get reported, instead of skipping the
     # tests.
-    import pyopencl as cl
-
-    import sys
     if len(sys.argv) > 1:
         exec(sys.argv[1])
     else:
         from py.test.cmdline import main
         main([__file__])
 
-# vim: filetype=pyopencl
+# vim: filetype=pyopencl:fdm=marker
